#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-09-19 11:35
Author:
    huayang (imhuay@163.com)
Subject:
    __init__
"""
from __future__ import annotations

# import os
# import sys
# import unittest

# from typing import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-09-18 20:12
Author:
    huayang (imhuay@163.com)
Subject:
    dataclass_utils
"""
from __future__ import annotations

# import os
# import sys
import unittest

# from typing import *

from huaytools import DataclassUtils


class TestDataclassUtils(unittest.TestCase):

    def test_get_field_names(self):
        from dataclasses import dataclass

        @dataclass
        class Demo:
            a: int
            b: str
            c = ...

        self.assertEqual(DataclassUtils.get_field_names(Demo), ['a', 'b'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-09-16 15:32
Author:
    huayang (imhuay@163.com)
Subject:
    dict_ex
"""
from __future__ import annotations

# import os
# import sys
# import unittest

from typing import Mapping, Union, Any, Iterable
from dataclasses import dataclass, fields

from huaytools.utils.dataclass_utils import DataclassUtils


class BunchDict(dict):
    """
    基于 dict 实现 Bunch 模式

    实现方法：
        - 通过重写 __getattr__、__setattr__、__delattr__ 同步 o.x 和 o['x'] 的行为
        - 为了防止与内部成员冲突，比如 __dict__，会预先调用 __getattribute__（优先级高于 __getattr__）

    Notes:
        - [2022.08.25] 通过对 __dict__ 添加 property 装饰限制修改，防止以下行为：
            - 如果直接向 __dict__ 添加属性，且存在同名 key，将导致 o.key 和 o['key'] 不一致；
            - 示例：
                ```python
                o = BunchDict(a=1, b=2)
                o.__dict__['a'] = 10
                print(o.a, o['a'])  # 10, 1
                ```

    Examples:
        >>> x = BunchDict(a=1, b=2)
        >>> x
        {'a': 1, 'b': 2}
        >>> x.c = 3  # x['c'] = 3
        >>> 'c' in x
        True
        >>> x['c']
        3
        >>> x['d'] = {'bar': 6}  # x.d = {'bar': 6}
        >>> hasattr(x, 'd')
        True
        >>> x.d.bar
        6
        >>> dir(x)
        ['a', 'b', 'c', 'd']
        >>> vars(x)
        {'a': 1, 'b': 2, 'c': 3, 'd': {'bar': 6}}
        >>> del x.a
        >>> x.a
        Traceback (most recent call last):
            ...
        AttributeError: a
        >>> x.__dict__
        {'b': 2, 'c': 3, 'd': {'bar': 6}}
        >>> x.__dict__ = {'a': 123}  # noqa
        Traceback (most recent call last):
            ...
        AttributeError: can't set attribute

        >>> y = {'foo': {'a': 1, 'bar': {'c': 'C'}}, 'b': 2}
        >>> y == BunchDict.from_dict(y) == BunchDict(y)
        True
        >>> x = BunchDict(y, d={'z': 26})
        >>> x.foo
        {'a': 1, 'bar': {'c': 'C'}}
        >>> x.foo.bar
        {'c': 'C'}
        >>> all(type(it) == BunchDict for it in [x.foo, x.foo.bar, x.d])  # noqa
        True

    References:
        - bunch（pip install bunch）
    """

    # __slots__ = ()
    __dict__ = property(lambda self: self)
    """ 禁止修改 __dict__ """

    def __dir__(self):
        """ 屏蔽其他属性或方法 """
        return self.keys()

    def __init__(self, seq: Union[Mapping, Iterable] = None, **kwargs):
        """"""
        super().__init__()  # 初始化 self，一个空 dict

        # 模拟向 dict 中添加元素的过程：https://docs.python.org/zh-cn/3/library/stdtypes.html#mapping-types-dict
        #   通过手动添加元素，确保每个类型为 dict 的值会被初始化为 BunchDict
        if seq is not None:
            if isinstance(seq, Mapping):
                seq = seq.items()
            for k, v in seq:
                self[k] = BunchDict.bunching(v)  # 如果 v 的类型为 dict，将被修改为 BunchDict

        for k, v in kwargs.items():
            self[k] = BunchDict.bunching(v)

    def __getattr__(self, name: str):
        """ 使 `o.name` 等价于 `o[name]` """
        try:
            # 除非是来自 object 或 dict 的特殊属性，比如 __doc__ 等
            return super().__getattribute__(name)
        except AttributeError:
            try:
                return self[name]
            except KeyError:
                raise AttributeError(name)

    def __setattr__(self, name: str, value):
        """ 使 `o.name = value` 等价于 `o[name] = value` """
        try:
            # 除非是来自 object 或 dict 的特殊属性，比如 __doc__ 等
            super().__getattribute__(name)
        except AttributeError:
            self[name] = value
        else:
            # setattr(self, name, value)  # RecursionError
            super().__setattr__(name, BunchDict.bunching(value))

    def __delattr__(self, name: str):
        """ 使 `del o.name` 等价于 `del o[name]` """
        try:
            # 除非是来自 object 或 dict 的特殊属性，比如 __doc__ 等
            super().__getattribute__(name)
        except AttributeError:
            try:
                del self[name]
            except KeyError:
                raise AttributeError(name)
        else:
            super().__delattr__(name)

    def __setitem__(self, key: str, value):
        """ 添加新元素，对 value 进行 bunching """
        super().__setitem__(key, BunchDict.bunching(value))

    @staticmethod
    def bunching(x: Union[Mapping, Any]) -> Union['BunchDict', Any]:
        return _bunching(x)

    @classmethod
    def from_dict(cls, d: dict) -> 'BunchDict':
        """ create from dict """
        return cls.bunching(d)

    def to_dict(self) -> dict:
        return _unbunch(self)


def _bunching(x) -> Union[BunchDict, Any]:
    """
    Recursively transforms a dictionary into a Bunch.

    Examples:
        >>> b = _bunching({'urmom': {'sez': {'what': 'what'}}})
        >>> b.urmom.sez.what
        'what'
        >>> b = _bunching({ 'lol': ('cats', {'hah':'i win'}), 'hello': [{'french':'salut', 'german':'hallo'}]})
        >>> b.hello[0].french
        'salut'
        >>> b.lol[1].hah
        'i win'
    """
    if isinstance(x, Mapping):
        return BunchDict((k, _bunching(v)) for k, v in x.items())
    elif isinstance(x, (list, tuple)):
        return type(x)(_bunching(v) for v in x)
    else:
        return x


def _unbunch(x: BunchDict) -> dict:
    """
    Recursively converts a Bunch into a dictionary.

    Examples:

        >>> b = BunchDict(foo=BunchDict(lol=True), hello=42, ponies='are pretty!')
        >>> _unbunch(b)
        {'foo': {'lol': True}, 'hello': 42, 'ponies': 'are pretty!'}

        unbunchify will handle intermediary dicts, lists and tuples (as well as
        their subclasses), but ymmv on custom datatypes.

        >>> b = BunchDict(foo=['bar', BunchDict(lol=True)], hello=42, ponies=('pretty!', BunchDict(lies='trouble!')))
        >>> _unbunch(b)
        {'foo': ['bar', {'lol': True}], 'hello': 42, 'ponies': ('pretty!', {'lies': 'trouble!'})}

        nb. As dicts are not hashable, they cannot be nested in sets/frozensets.
    """
    if isinstance(x, BunchDict):
        return dict((k, _unbunch(v)) for k, v in x.items())
    elif isinstance(x, (list, tuple)):
        return type(x)(_unbunch(v) for v in x)
    else:
        return x


@dataclass
class DataclassDict(dict):
    """
    Dataclass 字典：将 field 默认保存到一个字典中

    Examples:
        >>> from dataclasses import field
        >>> @dataclass
        ... class Features(DataclassDict):
        ...     a: int = 1
        ...     b: str = 'B'
        ...     c: list = field(default_factory=list)
        >>> f = Features(); f
        Features(a=1, b='B', c=[])
        >>> f.a = 2; f['a']
        2
        >>> f['a'] = 3; f.a
        3
        >>> f['d'] = 'D'
        >>> f.d  # noqa
        Traceback (most recent call last):
            ...
        AttributeError: 'Features' object has no attribute 'd'
        >>> not hasattr(f, 'd') and 'd' in f
        True
        >>> DataclassUtils.get_field_names(f)
        ['a', 'b', 'c']
        >>> f.c.append('Foo')
        >>> import json
        >>> json.dumps(f)  # 可以直接当做 dict 处理
        '{"a": 3, "b": "B", "c": ["Foo"], "d": "D"}'
    """

    def __post_init__(self):
        """"""
        # 把 field 依次添加到 dict 中
        for f in fields(self):
            self[f.name] = getattr(self, f.name)

    def __setattr__(self, key, value):
        """"""
        super().__setattr__(key, value)
        if key in DataclassUtils.get_field_names(self):
            super().__setitem__(key, value)

        # if key in DataclassUtils.get_field_names(self):
        #     super().__setattr__(key, value)
        #     super().__setitem__(key, value)
        # else:
        #     # raise KeyError(key)
        #     # 禁止添加新属性，除非是修改魔术属性
        #     try:
        #         super().__getattribute__(key)
        #     except AttributeError:
        #         raise KeyError(key)
        #     else:
        #         super().__setattr__(key, value)

    def __setitem__(self, key, value):
        """"""
        super().__setitem__(key, value)
        if key in DataclassUtils.get_field_names(self):
            super().__setattr__(key, value)

        # if key in DataclassUtils.get_field_names(self):
        #     super().__setitem__(key, value)
        #     super().__setattr__(key, value)
        # else:
        #     raise KeyError(key)

    # def __str__(self):
    #     return dict.__str__(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-09-06 10:57
Author:
    huayang (imhuay@163.com)
Subject:
    python_utils
"""
from __future__ import annotations

import os
# import sys
import inspect
import sys

from typing import Tuple, List


class PythonUtils:
    """"""

    @staticmethod
    def get_version() -> Tuple[int]:
        """

        """
        return sys.version_info[:3]  # (3, 9, 13)

    @staticmethod
    def get_frame(stack_level: int = 0):
        """
        Examples:
            >>> def foo():
            ...     f = PythonUtils.get_frame()
            ...     return f.f_code.co_name
            >>> foo()
            'foo'
        """
        assert stack_level >= 0
        frame = inspect.currentframe()  # 获取当前 frame，即 `get_frame` 的 frame
        for _ in range(stack_level + 1):  # 向上逐级回溯调用者的 frame，`+1` 表示跳过 get_frame 本身的调用
            frame = frame.f_back

        return frame

    @staticmethod
    def get_caller_name(stack_level: int = 1) -> str:
        """
        获取调用者的名字

        使用场景：
            假设 foo 内部调用了 get_caller_name，则 get_caller_name 默认（level=1）会返回调用了 foo 的对象名；
            如果调用 foo 的对象是一个方法/模块/类，则返回该相应的的方法名/模块名/类名；

        Args:
            stack_level: 回溯层级，默认为 1；

        Examples:
            >>> def foo():
            ...     name = PythonUtils.get_caller_name(0)  # name of who called `get_caller_name`
            ...     print(name)
            >>> foo()
            foo
            >>> def foo():
            ...     name = PythonUtils.get_caller_name()  # name of who called `foo`
            ...     print(name)
            >>> def bar():
            ...     foo()
            >>> bar()
            bar
            >>> def zoo():
            ...     foo()
            >>> zoo()
            zoo
        """
        assert stack_level >= 0
        frame = PythonUtils.get_frame(stack_level + 1)
        co_name = frame.f_code.co_name

        # 当调用方是一个模块，此时返回模块的文件名
        if co_name == '<module>':
            return os.path.basename(frame.f_code.co_filename)

        return co_name

    @staticmethod
    def get_lineno(stack_level: int = 0) -> int:
        """
        获取调用时行号

        Args:
            stack_level: 回溯层级

        Examples:
            # example1: default
            lno = get_lineno()  # this line
            print(lno)

            # example2: stack_level > 0
            def foo():
                return get_lineno(1)

            lno = foo()  # this line
            print(lno)
        """
        return PythonUtils.get_frame(stack_level + 1).f_lineno

    @staticmethod
    def get_annotation_names(obj: object) -> List[str]:
        return list(obj.__annotations__.keys())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-08-29 13:13
Author:
    huayang (imhuay@163.com)
Subject:
    __init__.py
"""

from .utils import *  # noqa
from .extensions import *  # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-08-30 10:28
Author:
    huayang (imhuay@163.com)
Subject:
    _common
"""
from __future__ import annotations

import time
import logging
import functools

from typing import Callable

from .python_utils import PythonUtils

logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s@%(lineno)dL : %(message)s',
                    datefmt='%Y.%m.%d %H:%M:%S',
                    level=logging.INFO)
"""
References: https://docs.python.org/zh-cn/3/library/logging.html#logrecord-attributes
"""


class CommonUtils:
    """"""

    @staticmethod
    def get_logger(name: str = None):
        """返回一个 logger"""
        return logging.getLogger(name or PythonUtils.get_caller_name())


get_logger = CommonUtils.get_logger


class CodeTimer:
    name: str
    msg_start: str
    msg_end: str
    start: float
    stream: Callable

    def __init__(self, name: str = '', stream: Callable = print):
        """
        Args:
            name:
            stream:

        Examples:
            import time

            @code_timer()
            def foo():
                time.sleep(1)

            with code_timer('test'):
                time.sleep(1)

        TODO: add color
        """
        self.name = name
        # self.msg_start = msg_start or 'Start {name}{{'
        # self.msg_end = msg_end or '}} End - Spend {cost:.5f}s. \n'
        self.stream = stream

    def __call__(self, func):
        """"""
        if not self.name:
            self.name = func.__name__

        @functools.wraps(func)
        def wrap(*args, **kwargs):
            """"""
            self.info_start(self.name)
            self.start = time.time()
            func(*args, **kwargs)
            self.info_end()

        return wrap

    def info_start(self, name=''):
        if name:
            name = f' {name}'
        line = f' @{PythonUtils.get_lineno(2)}L'
        self.stream(f'Start{name}{line}{{')

    def info_end(self):
        cost = time.time() - self.start
        self.stream(f'}} End - Spend {cost:.5f}s. \n')

    def __enter__(self):
        self.info_start(self.name)
        self.start = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.info_end()


code_timer = CodeTimer()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-09-13 16:20
Author:
    huayang (imhuay@163.com)
Subject:
    abc_ex
"""
from __future__ import annotations

# import os
# import sys
# import unittest
import abc

# from typing import *
from threading import Lock


class SingletonMeta(type):
    """
    实现单例模式的元类（线程安全）

    Examples:
        >>> class A(metaclass=SingletonMeta): pass
        >>> A() is A()
        True
    """
    _instances = dict()
    _lock = Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:  # 第一次判断，防止每次判断都执行锁操作
            with cls._lock:
                if cls not in cls._instances:  # 第二次判断：防止实例被多次创建
                    # import time
                    # time.sleep(1)  # test thread safety
                    cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class SingletonABC(abc.ABC):
    """
    实现单例模式的抽象类（线程安全）

    Notes:
        与 SingletonMeta 的区别：
            SingletonMeta 对单例是否存在的判断发生在 __init__ 之前，
            而 SingletonABC 发生在 __init__ 之后；
        考虑以下代码：

        class A(metaclass=SingletonMeta):
            def __init__(self, v):
                self.v = v

        class B(SingletonABC):
            def __init__(self, v):
                self.v = v

        a1, a2, a3 = A(1), A(2), A(3)
        b1, b2, b3 = B(1), B(2), B(3)
        print(a1 is a2 is a3)  # True
        print(b1 is b2 is b3)  # True
        print(a3.v)  # 1，新的 v 未生效，说明 __init__ 没有执行
        print(b1.v)  # 3，新的 v 生效了，可见已经执行了 __init__

    Examples:
        >>> class A(SingletonABC): pass
        >>> A() is A()
        True
    """
    _instance: SingletonABC
    _lock = Lock()

    def __new__(cls, *args, **kwargs):
        # try:
        #     return cls._instance
        # except AttributeError:
        #     cls._instance = super().__new__(cls)
        #     return cls._instance
        if not hasattr(cls, '_instance'):  # # 第一次判断，防止每次判断都执行锁操作
            with cls._lock:
                if not hasattr(cls, '_instance'):  # 第二次判断：防止实例被多次创建
                    # import time
                    # time.sleep(1)  # test thread safety
                    cls._instance = super().__new__(cls)
        return cls._instance



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-09-13 11:42
Author:
    huayang (imhuay@163.com)
Subject:
    __init__.py
"""

from .abc_ex import (
    SingletonMeta,
    SingletonABC
)
from .dict_ex import (
    BunchDict,
    DataclassDict
)

__all__ = [
    # abc_ex
    'SingletonMeta',
    'SingletonABC',
    # dict_ex
    'BunchDict',
    'DataclassDict'
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-09-16 17:07
Author:
    huayang (imhuay@163.com)
Subject:
    dict_ex
"""
from __future__ import annotations

import json
# import os
# import sys
import unittest

# from typing import *
from dataclasses import dataclass, fields, field

from huaytools.extensions import BunchDict, DataclassDict


class TestBunchDict(unittest.TestCase):

    def test_base(self):
        x = BunchDict(a=1, b=2)
        self.assertEqual(x, {'a': 1, 'b': 2})

        x.c = 3
        self.assertIn('c', x)
        self.assertEqual(x['c'], x.c)

        x['d'] = {'foo': 6}
        self.assertIs(type(x.d), BunchDict)
        self.assertTrue(hasattr(x, 'd'))
        self.assertTrue(hasattr(x.d, 'foo'))

        x.c = {'bar': 8}
        self.assertIs(type(x.c), BunchDict)

        self.assertEqual(dir(x), ['a', 'b', 'c', 'd'])
        self.assertEqual(vars(x), {'a': 1, 'b': 2, 'c': {'bar': 8}, 'd': {'foo': 6}})
        self.assertEqual(str(x), "{'a': 1, 'b': 2, 'c': {'bar': 8}, 'd': {'foo': 6}}")

        with self.assertRaises(AttributeError):
            del x.a
            getattr(x, 'a')

        with self.assertRaises(AttributeError):
            x.__dict__ = {}  # noqa

        x.__doc__ = {'a': 1}
        self.assertIs(type(x.__doc__), BunchDict)

    def test_from_dict(self):
        y = {'foo': {'a': 1, 'bar': {'c': 'C'}}, 'b': 2}
        x = BunchDict.from_dict(y)
        self.assertEqual(y, x)
        self.assertEqual(x, BunchDict(y))

        x = BunchDict(y, d={'z': 26})
        self.assertEqual(x, {'foo': {'a': 1, 'bar': {'c': 'C'}}, 'b': 2, 'd': {'z': 26}})
        self.assertEqual(x.foo, {'a': 1, 'bar': {'c': 'C'}})
        self.assertEqual(x.foo.bar, {'c': 'C'})
        for it in [x.foo, x.foo.bar, x.d]:
            self.assertIs(type(it), BunchDict)

    def test_to_dict(self):
        y = {'foo': {'a': 1, 'bar': {'c': 'C'}}, 'b': 2}
        x = BunchDict(y).to_dict()
        self.assertEqual(x, y)
        self.assertIs(type(x), dict)


class TestDataclassDict(unittest.TestCase):

    @dataclass
    class Features(DataclassDict):
        a: int
        b: str = 'B'
        c: list = field(default_factory=list)

    def test_base(self):
        f = self.Features(a=1)
        self.assertTrue(f.a == f['a'] == 1)
        self.assertEqual(json.dumps(f), '{"a": 1, "b": "B", "c": []}')
        self.assertEqual(f, {'a': 1, 'b': 'B', 'c': []})
        self.assertEqual(vars(f), {'a': 1, 'b': 'B', 'c': []})

        field_names = [i.name for i in fields(f)]
        self.assertEqual(field_names, ['a', 'b', 'c'])

        f.a = ten = 10
        self.assertEqual(f.a, ten)
        self.assertEqual(f['a'], ten)

        f['b'] = bar = 'Bar'
        self.assertEqual(f.b, bar)
        self.assertEqual(f['b'], bar)

        f['d'] = 42
        with self.assertRaises(AttributeError):
            getattr(f, 'd')

        f.pop('d')
        f.c.append('Foo')
        self.assertEqual(json.dumps(f), '{"a": 10, "b": "Bar", "c": ["Foo"]}')

        f.__doc__ = 'doc'
        self.assertEqual(f.__doc__, 'doc')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-09-06 11:04
Author:
    huayang (imhuay@163.com)
Subject:
    python_utils
"""
from __future__ import annotations

import inspect
import unittest

from huaytools import PythonUtils


class TestPythonUtils(unittest.TestCase):
    """"""

    def test_get_version(self):
        v = PythonUtils.get_version()
        self.assertEqual(3, len(v))
        self.assertGreaterEqual(3, v[0])

    def test_get_frame(self):
        frame = PythonUtils.get_frame()
        self.assertEqual(self.test_get_frame.__name__, frame.f_code.co_name)

        frame2 = inspect.currentframe()
        self.assertTrue(frame2 is frame)

        frame3 = PythonUtils.get_frame(1)
        self.assertTrue(frame3 is frame.f_back)

    def test_get_caller_name(self):
        def foo():
            caller_name = PythonUtils.get_caller_name()
            return caller_name

        name = foo()
        self.assertEqual(self.test_get_caller_name.__name__, name)

        name = PythonUtils.get_caller_name(0)
        self.assertEqual(self.test_get_caller_name.__name__, name)

    def test_get_lineno(self):
        expected_lno = 50
        lno = PythonUtils.get_lineno()
        self.assertEqual(expected_lno, lno)

        def foo():
            return PythonUtils.get_lineno(1)

        lno = foo()
        self.assertEqual(expected_lno + 6, lno)

    def test_get_annotation_names(self):

        class Demo:
            a: int
            b: str
            c = []

        names = PythonUtils.get_annotation_names(Demo)
        self.assertEqual(names, ['a', 'b'])

        names = PythonUtils.get_annotation_names(Demo())
        self.assertEqual(names, ['a', 'b'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-09-05 17:12
Author:
    huayang (imhuay@163.com)
Subject:
    _common
"""
from __future__ import annotations

from unittest import TestCase

from huaytools import CodeTimer, get_logger


class TestCommonUtils(TestCase):  # noqa

    def test_get_logger(self):

        logger = get_logger()
        self.assertEqual(self.test_get_logger.__name__, logger.name)

        logger2 = get_logger(logger.name)
        self.assertTrue(logger2 is logger)


class TestCodeTimer(TestCase):

    def test_base(self):
        logger = get_logger()

        def func():
            pass

        CodeTimer(stream=logger.info)(func)()

        with CodeTimer(stream=logger.info):
            func()

        self.assertTrue(True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-08-29 14:12
Author:
    huayang (imhuay@163.com)
Subject:
    __init__.py
"""

from ._common import (
    CommonUtils,
    get_logger,
    code_timer,
    CodeTimer
)
from .python_utils import PythonUtils
from .dataclass_utils import DataclassUtils

__all__ = [
    # _common
    'CommonUtils',
    'get_logger',
    'code_timer',
    'CodeTimer',
    # python_utils
    'PythonUtils',
    # dataclass_utils
    'DataclassUtils'
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-09-16 20:09
Author:
    huayang (imhuay@163.com)
Subject:
    dataclass_utils
"""
from __future__ import annotations

# import os
# import sys
# import unittest

# from typing import *
from dataclasses import fields


class DataclassUtils:

    @staticmethod
    def get_field_names(class_or_instance):
        return [f.name for f in fields(class_or_instance)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time:
    2022-09-13 16:22
Author:
    huayang (imhuay@163.com)
Subject:
    meta_ex
"""
from __future__ import annotations

# import os
# import sys
# import json
import unittest

# from typing import *
from concurrent.futures import ThreadPoolExecutor, as_completed

from huaytools.extensions import SingletonMeta
from huaytools.extensions import SingletonABC


# from huaytools.utils import get_logger
# logger = get_logger()


class TestSingletonMeta(unittest.TestCase):
    class A(metaclass=SingletonMeta):
        def __init__(self, value=None):
            self.value = value

    def test_base(self):
        self.assertIs(self.A(), self.A())

    def test_thread_safe(self):
        """
        记录一个测试问题：
            在测试线程不安全时（通过 time.sleep），单独运行这条测试会得到期望的结果（期望不通过）；
            但是直接运行整个文件时，会得到相反的结果；
        """
        instances = []
        n_instance = 100
        with ThreadPoolExecutor(max_workers=10) as executor:
            todo = []
            for i in range(n_instance):  # 模拟多个任务
                future = executor.submit(self.A, value=i)
                todo.append(future)

            for future in as_completed(todo):  # 并发执行
                instances.append(future.result())

        # print(instances[:3])
        for i in instances:
            self.assertIsNotNone(i)

        singleton = instances[0]
        for i in instances:
            self.assertIs(i, singleton)

    # def test_time_cost(self):
    #     """
    #     测试显示加入两次判断，是否能够提升速度
    #     """
    #     import time
    #     from threading import Lock
    #
    #     class _SingletonMeta(type):
    #         _instances = dict()
    #         _lock = Lock()
    #
    #         def __call__(cls, *args, **kwargs):
    #             # 去除第一次判断
    #             # if cls not in cls._instances:
    #             with cls._lock:
    #                 if cls not in cls._instances:  # 第二次判断：防止实例被多次创建
    #                     # import time
    #                     # time.sleep(1)  # test thread safety
    #                     cls._instances[cls] = super().__call__(*args, **kwargs)
    #             return cls._instances[cls]
    #
    #     class A(metaclass=SingletonMeta):
    #         pass
    #
    #     class B(metaclass=_SingletonMeta):
    #         pass
    #
    #     n_time = 100000
    #     tmp = []
    #     start = time.time()
    #     for _ in range(n_time):
    #         tmp.append(A())
    #     t1 = time.time() - start
    #     # print()
    #     # print(t1)
    #
    #     tmp = []
    #     start = time.time()
    #     for _ in range(n_time):
    #         tmp.append(B())
    #     t2 = time.time() - start
    #     # print(t2)
    #     self.assertLess(t1, t2)


class TestSingletonABC(TestSingletonMeta):
    class A(SingletonABC):
        def __init__(self, value=None):
            self.value = value

    def test_base(self):
        super().test_base()

    def test_thread_safe(self):
        super().test_thread_safe()

    def test_time_cost(self):
        pass
